---------------------------------------------------------------------
---------------- Functions and hook definitions  --------------------
---------------------------------------------------------------------

-- local function rotateTheta(theta, delta, theta_prime) 
--     -- local increment = FixedDiv(AngleFixed(theta_prime), AngleFixed(theta))
    
--     -- local new_theta_value = FixedMul(increment, theta) / delta

--     -- return new_theta_value



-- end

--Returns whether the car can bump the collided into object
local function CheckBump(car_mo, collide_mo)
    return car_mo.player.speed > BUMP_SPEED and collide_mo.flags & BUMP_MF and L_ZCollide(car_mo, collide_mo)
end


local function TireEffects(player)

    if(P_IsObjectOnGround(player.mo)) then
        
        if(player.mo.applied_speed > 1) then
            SpawnOnSide(player, MT_TRCK)    
        end

        --Drifting effects
        if(player.braking_input >= 1 and player.forward_input >= 1) then
            SpawnOnSide(player, MT_SPINDUST)
        end

        if(player.mo.applied_speed > CAR_ACCELERATION) then 
            SpawnOnSide(player, MT_FIRE)
        end
    end 
end

local function CarSounds(player)

    --Using the horn
    if(input.gameControlDown(GC_CUSTOM3) and not S_SoundPlaying(player.mo, sfx_horn)) then
        S_StartSound(player.mo, sfx_horn)
    end

    --Starting the engine
    if(not S_SoundPlaying(player.mo, sfx_rev) and player.speed == 0 and player.mo.applied_speed >= 1) then
        S_StartSound(player.mo, sfx_rev)
    end
    
    --Idle working engine sound
    if(player.sound_tics_idle <= 0 and player.speed <= IDLE_SPEED) then
        S_StartSound(player.mo, sfx_idle)
        player.sound_tics_idle = MAX_IDLE_SOUND_TICS
    else
        player.sound_tics_idle = $-1
    end

    --Mid working engine sound
    if(player.sound_tics_mid <= 0 and player.speed <= MID_SPEED and player.speed > IDLE_SPEED) then
        S_StartSound(player.mo, sfx_mid)
        player.sound_tics_mid = MAX_MID_SOUND_TICS
        S_StopSoundByID(player.mo, sfx_high)
    else
        player.sound_tics_mid = $-1
    end

    --High working engine sound
    if(player.sound_tics_high <= 0 and player.speed > MID_SPEED) then
        S_StartSound(player.mo, sfx_high)
        player.sound_tics_high = MAX_HIGH_SOUND_TICS
        S_StopSoundByID(player.mo, sfx_mid)
    else
        player.sound_tics_high = $-1
    end

    -- Braking sound
    if(P_IsObjectOnGround(player.mo) and player.forward_input > 0 and player.braking_input > 0 and not S_SoundPlaying(player.mo, sfx_brake)) then
        S_StartSound(player.mo, sfx_brake)
    elseif(player.braking_input <= 0) then
        S_StopSoundByID(player.mo, sfx_brake)
    end

end

local function CarPreThink(player)

    --Cancel jumping and regular movement
    player.cmd.forwardmove = 0
    player.cmd.sidemove = 0
    player.pflags = $|PF_JUMPSTASIS
    
    --The finalized force and angle
    -- player.mo.applied_speed = 0
    -- player.mo.applied_angle = 0

    --Side input between -1000 and 1000
    local side_input = 0
    local steering = 0

    local side_input_keyboard = (input.gameControlDown(GC_STRAFELEFT) - input.gameControlDown(GC_STRAFERIGHT)) * 1000
    local side_input_joystick = -clamp(input.joyAxis(JA_STRAFE), -1000, 1000)
    
    -- Choosing either the keyboard or joystick input for pitch
    if(abs(side_input_joystick) >= abs(side_input_keyboard)) then
        side_input = side_input_joystick 
    else 
        side_input = side_input_keyboard             
    end

    -- Car movement inputs
    player.forward_input = max(clamp(player.cmd.buttons & BT_JUMP, 0, 1), input.gameControlDown(GC_FIRE) and 1 or 0)
    player.braking_input = max((player.spinheld >= 1) and 1 or 0, input.gameControlDown(GC_FIRENORMAL) and 1 or 0)
    

    --The maximum applied acceleration
    local max_acceleration = (player.powers[pw_sneakers] == 0) and CAR_ACCELERATION or CAR_ACCELERATION_BOOSTED
    --The maximum speed 
    local max_speed = (player.powers[pw_sneakers] == 0) and SPEED_LIMIT or SPEED_LIMIT_BOOSTED


    --Apply additional special boost to maximum acceleration and speeds
    if(player.special_boost_tics > 0) then --Special speed boost is in the affect
        max_speed = $ + SPECIAL_BOOST_CAR_SPEED
        max_acceleration = $ + SPECIAL_BOOST_CAR_ACCEL
    end

    --How strong is the brake power
    local brake_power = 2*max_acceleration/3
    --One angular percent of maximum steering
    local angle_unit = FixedAngle(FixedDiv(AngleFixed(STEERING_LIMIT_ANGLE.MAX), 1000*FRACUNIT)) 
    -- The angle caused by steerig relative to the current angle of the car
    local steering_angle = player.mo.angle + angle_unit * side_input
    --Player's desired force to move the car forward
    local forward_force = player.forward_input*max_acceleration
    --Player's desired force that stops the car's movement
    local braking_force = player.braking_input*brake_power
    --Total velocity 
    local desired_velocity = 0

    --Forward and backward speed (positive and negative respectively)
    local signed_speed = FixedMul(player.mo.momx, cos(player.mo.angle)) + FixedMul(player.mo.momy, sin(player.mo.angle))

    --Make sure backward movement is not too strong
    if(signed_speed <= 0) then
        braking_force = 0
    end
    --Regular movement
    desired_velocity = forward_force - braking_force

    --Friction changes
    local sector_friction = player.mo.subsector.sector.friction
    local applied_friction = sector_friction+FRACUNIT/20

    --Liquid effects reducing the speed
    if(player.mo.eflags & MFE_UNDERWATER) then
        player.mo.friction = player.mo.subsector.sector.friction-(FRACUNIT/5)
    else
        -- Regular friction
        player.mo.friction = clamp(applied_friction, sector_friction, FRACUNIT-1)
    end

    --Limiting airbone speed
    if(not P_IsObjectOnGround(player.mo)) then
        desired_velocity = clamp($, -1, 1)
    end
    
    --Limiting the speed 
    if(player.speed > max_speed) then
        desired_velocity = 0
    end


    --Final Force vectory to be applied
    player.mo.applied_angle = steering_angle
    player.mo.applied_speed = desired_velocity



    --Making sure that autobrake feature woudn't slow us down while driving
    if(abs(desired_velocity) > FRACUNIT or not P_IsObjectOnGround(player.mo)) then
        player.cmd.forwardmove = 1
    end

    --Displays tire effects
    TireEffects(player)

    --Executes the car sounds
    CarSounds(player)
end

local function CarPlayerThink(player)

    --The point: turn the car during the steering turn depending on the speed of the car, faster the slower the turn
    --The steering angle not relative to the player's angle
    local desired_steering_angle = player.mo.applied_angle - player.mo.angle
    local max_steering_speed = REGULAR_STEERING_SPEED

    -- Change some behaviors while braking
    if(player.braking_input >= 1) then
        -- Increase the steering angle to simulate "drifting"
        max_steering_speed = DRIFTING_STEERING_SPEED
    end

    --The maximum speed 
    local max_speed = (player.powers[pw_sneakers] == 0) and SPEED_LIMIT or SPEED_LIMIT_BOOSTED

    --Applying the steering
    local turn_speed = clamp(FixedDiv(clamp(player.speed, 0, max_speed), max_speed), 0, max_steering_speed)
    local final_steering_angle = ease.linear(turn_speed, 0, desired_steering_angle)
    player.mo.angle = $ + final_steering_angle

    
    if(player.mo.standingslope ~= nil) then
        -- print("A: "..player.mo.standingslope.zangle/ANG1)
        -- print("D: "..player.mo.standingslope.zdelta)
    end

    --[[
    --Driving off the ledge
    if(player.mo.standingslope ~= nil and not P_IsObjectOnGround(player.mo)) then
        P_SetObjectMomZ(player.mo, 10*FRACUNIT, true)
        print("Jump!")
        -- P_Thrust(player.mo, player.mo.angle, player.speed/10)
    end
    ]]--
end

local function CarPostThink(player)
    -- local movement_angle = R_PointToAngle2(0, 0, player.mo.momx, player.mo.momy)
    -- camera.angle = $ + ease.linear(FRACUNIT, 0, movement_angle)

    -- Car movement
    if(P_IsObjectOnGround(player.mo)) then
        P_Thrust(player.mo, player.mo.angle, player.mo.applied_speed)
    end

    
    
end


local function CarMobjMoveCollide(car_mo, collide_mo)

    -- Collide with the object only if it meets criteria
    if(CheckBump(car_mo, collide_mo)) then

        S_StartSound(car_mo, sfx_bump)
        --Destroying the monitors
        if(collide_mo.flags & MF_MONITOR) then
            P_DamageMobj(collide_mo, car_mo, car_mo, 1)
         
        --Getting enemies out of the way
        elseif(collide_mo.flags & MF_ENEMY) then    
        
            -- Bumping objects out of the way
            P_SetObjectMomZ(collide_mo, FRACUNIT*10, false)
            P_InstaThrust(collide_mo, R_PointToAngle2(0, 0, car_mo.momx, car_mo.momy), -FRACUNIT*5)
        end
       
    end
end


local function CarMobjDamage(car_mo, inflictor, source, damage, damagetype)
    --Evading damage while bumping the enemy
    -- if (inflictor == source and damage == 1 -- If light damage if source damaged us
    -- or (source ~= nil and source & BUMP_MF)) then  -- If source is bumpable
    if(inflictor ~= nil and inflictor == source and CheckBump(car_mo, inflictor)) then
        return true --Evade damage
    end
end


---------------------------------------------------------------------
---------- Connecting all of the hooks from this file  --------------
---------------------------------------------------------------------


addHook("PlayerThink", function(player) 
    if(Valid(player.mo, "jazz") and player.tf_mode == TF_MODES.CAR_MODE) then
        CarPlayerThink(player) 
    end
end)



--The Base Thinker that plays before others,
--Mostly used to record players input
addHook("PreThinkFrame", function()
	for player in players.iterate() do
        if(Valid(player.mo, "jazz") and player.tf_mode == TF_MODES.CAR_MODE) then
            CarPreThink(player)
        end
    end
end)


--The Base Thinker that plays before others,
--Mostly used to record players input
addHook("PostThinkFrame", function()
	for player in players.iterate() do
        if(Valid(player.mo, "jazz") and player.tf_mode == TF_MODES.CAR_MODE) then
            CarPostThink(player)
        end
    end
end)

--Handles collision of the car mode with the objects
addHook("MobjMoveCollide", function(mo, collide_mo)
    if(Valid(mo, "jazz") and collide_mo ~= nil and mo.player.tf_mode == TF_MODES.CAR_MODE) then
        CarMobjMoveCollide(mo, collide_mo)
    end
end, MT_PLAYER)


addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
    
    if(Valid(target, "jazz") and target.player.tf_mode == TF_MODES.CAR_MODE) then
        return CarMobjDamage(target, inflictor, source, damage, damagetype)
    end
end, MT_PLAYER)