---------------------------------------------------------------------
------------------------ All freeslot Extra  ------------------------
---------------------------------------------------------------------


--Sprites
freeslot("SPR2_CAR_", "SPR2_TFCA", "SPR2_TFBO", "SPR_BLST", "SPR_BULL", "SPR_TRCK", "SPR2_BUSN", "SPR2_BUJU")
--States
freeslot("S_CAR_MODE", "S_TF_TO_BOT", 
"S_TF_TO_CAR", "S_BLAST_SHOT", "S_BLAST_EXPLODE", "S_BLAST_DUST", "S_BULL_SHOT", "S_WEAPON_STANCE",
 "S_CLIMBING", "S_TRCK", "S_FIRE")

--Sounds
freeslot("sfx_tf_car", "sfx_tf_bot", "sfx_rev", "sfx_idle", "sfx_mid", "sfx_high", "sfx_brake", "sfx_bump", "sfx_horn", 
"sfx_bullet", "sfx_hit", "sfx_miss", "sfx_blast", "sfx_expld", "sfx_land", "sfx_climb",
 "sfx_begn01", "sfx_begn02", "sfx_begn03", "sfx_begn04", "sfx_begn05", "sfx_win01", "sfx_win02", "sfx_win03", "sfx_win04", "sfx_win05")

--Objects
freeslot("MT_BLST", "MT_BULL", "MT_TRCK", "MT_FIRE", "MT_BDST")


------------------------------------------------------------------------------------
------------------------ All rawset Variables and Functions ------------------------
------------------------------------------------------------------------------------
--- NOTE: Constant variables are FULLY CAPITALIZED (don't change constants!!!) -----
------------------------------------------------------------------------------------


--------------- Any Mode Constants ---------------

--Transformation
rawset(_G, "TF_MODES", {BOT_MODE = 0, CAR_MODE = 1}) -- Simulates Enums for all possible transformation modes
rawset(_G, "TRANFORMATION_TICS_MAX", TICRATE/4) --The cooldown of the transformation


--Special Speed Boosts for bot mode
rawset(_G, "SPECIAL_BOOST_TICS_MAX", 5*TICRATE) -- Maximum speedup time by the special ability
rawset(_G, "SPECIAL_BOOST_RING_TICS", 9) -- How much tics added to special speed boost timer when picking up a ring
rawset(_G, "SPECIAL_BOOST_DMG_TICS", 25)
rawset(_G, "SPECIAL_BOOST_MONITOR_TICS", 15)
rawset(_G, "SPECIAL_BOOST_SPRING_TICS", 25)
rawset(_G, "SPECIAL_BOOST_BOT_SPEED", 6*FRACUNIT) -- The amount of speedup
rawset(_G, "SPECIAL_BOOST_BOT_ACCEL", 20) -- The amount of increased acceleration
rawset(_G, "SPECIAL_BOOST_CAR_SPEED", 15*FRACUNIT) -- For car mode
rawset(_G, "SPECIAL_BOOST_CAR_ACCEL", 1*FRACUNIT) -- For car mode

rawset(_G, "BREAK_SPEED", 40*FRACUNIT) -- At what minimum speed will Jazz be able to break walls, ceiling, floors, and empower springs

--------------- Car Mode constants ---------------


--Speed related
rawset(_G, "CAR_ACCELERATION", 4*FRACUNIT) --Force applied every tic
rawset(_G, "CAR_ACCELERATION_BOOSTED", CAR_ACCELERATION*2) --Force applied every tic while using speed sneakers
rawset(_G, "SPEED_LIMIT", 65*FRACUNIT) --Fastest speed while driving normally
rawset(_G, "SPEED_LIMIT_BOOSTED", 3*SPEED_LIMIT/2) --Fastest speed while driving while using speed sneakers
rawset(_G, "IDLE_SPEED", FRACUNIT) --Up to what speed the engine is in the idle state
rawset(_G, "MID_SPEED", SPEED_LIMIT/2) --Up to what speed the engine is in the middle speed state
rawset(_G, "STEERING_LIMIT_ANGLE", {MAX = ANGLE_45, MIN = -ANGLE_45}) --Furthest steering angle
rawset(_G, "REGULAR_STEERING_SPEED", FRACUNIT/10) --How fast can you steer regularly
rawset(_G, "DRIFTING_STEERING_SPEED", FRACUNIT/7) --How fast can you steer while drifting

--Sound and Voice related
rawset(_G, "MAX_IDLE_SOUND_TICS", 34)
rawset(_G, "MAX_MID_SOUND_TICS", 47)
rawset(_G, "MAX_HIGH_SOUND_TICS", 26)
rawset(_G, "VOICE_VOLUME", 255)

--Bumping objects out of hte way
rawset(_G, "BUMP_MF", MF_ENEMY|MF_MONITOR) --Objects that are able to be bumped
rawset(_G, "BUMP_NOT_MF", 0) -- Not used, but intended to detect objects that can't be bumped
rawset(_G, "BUMP_SPEED", MID_SPEED) --At what speed car mode can bump objects

--Misc
rawset(_G, "CAR_SHIELD_SCALE", FRACUNIT)


--------------- Bot Mode Constants ---------------


--Weapons
rawset(_G, "BULLET_TICS_MAX", 10) --Total period of bullets being shot (not every tic of it)
rawset(_G, "BLAST_KNOCKBACK", 20*FRACUNIT)

--Transformation into bot
rawset(_G, "TO_BOT_JUMP", FRACUNIT*10) --Jump force applied whlie transforming into bot 

--Climbing (maybe deprecated)
rawset(_G, "CLIMB_WALL_DISTANCE", 30*FRACUNIT) --
rawset(_G, "CLIMBING_SPEED", 5*FRACUNIT) --Speed of climbing all four directions



--------------- Functions Mode Constants ---------------


--Clamps the integer num between inlcusive bounds
rawset(_G, "clamp", function(num, lower, upper)
	if(num > upper) then
		num = upper
	elseif(num < lower) then
		num = lower
	end
	return num
	-- return max(lower, min(upper, num))
end)


--Checks whether the mobject is valid and (optionally) has the correct skin 
rawset(_G, "Valid", function(mo, skin)
	return mo ~= nil and mo.valid == true and mo.skin == skin and mo.state ~= S_NULL and mo.player.playerstate == PST_LIVE and mo.player.exiting == 0--and mo.state ~= states[mo.state].deathstate
end)

--Gets the actual absolute speed in 3D space at any angle
rawset(_G, "GetRealSpeed", function(player)
    local vs = abs(player.mo.momz) -- Vertical speed
    local hs = player.speed -- Horizontal speed
	return FixedSqrt(FixedMul(vs, vs) + FixedMul(hs, hs))
	-- return max(lower, min(upper, num))
end)


--Function mostly shown by clairebun to use on SRB2 discord page
--mo1 (mobj_t): first mobj to check for collision
--mo2 (mobj_t): second mobj to check for collision
--extraheight (int): extra height added to the mo2
--returns: true if two mobjects collide vertically, false otherwise
rawset(_G, "L_ZCollide", function(mo1, mo2, extraheight)
	if(extraheight == nil) then
		extraheight = 0
	end
    if mo1.z > mo2.height+mo2.z+extraheight then return false end
    if mo2.z > mo1.height+mo1.z then return false end
    return true
end)

--Spawns the objects on the side of the player
--Returns right and left spawned objects
rawset(_G, "SpawnOnSide", function(player, toSpawn)
    local right_x = player.mo.x + FixedMul(skins[player.mo.skin].radius, InvAngle(cos(player.mo.angle+ANGLE_90)))
    local right_y = player.mo.y + FixedMul(skins[player.mo.skin].radius, InvAngle(sin(player.mo.angle+ANGLE_90)))
    local left_x  = player.mo.x + FixedMul(skins[player.mo.skin].radius, InvAngle(cos(player.mo.angle-ANGLE_90)))
    local left_y  = player.mo.y + FixedMul(skins[player.mo.skin].radius, InvAngle(sin(player.mo.angle-ANGLE_90)))
    
    local z = player.mo.z
    if(player.mo.eflags & MFE_VERTICALFLIP) then
        z = $ + player.mo.height - mobjinfo[toSpawn].height
    end

    --Spawn at the side
    local right = P_SpawnMobj(right_x, right_y, z, toSpawn)
    local left = P_SpawnMobj(left_x, left_y, z, toSpawn)

    return right, left
end)


--Switch to the jazz skin
--Returns true if could switch to jazz, false if player was already it
rawset(_G, "SwitchTo", function(player)
	if(Valid(player.mo)) then
		return false -- do not switch to jazz if the player is already jazz
	end
    
	--Set controls
	player.jumpheld = 0
	player.spinheld = 0
    player.forwardheld = 0
	player.prevjumpheld = 0
	player.prevspinheld = 0

    player.mo.prevstate = player.mo.state
	
    --The previous vertical momentum
    player.mo.prevmomz = 0
	
    --Is jump special available?
    player.jumpspecial = 0 -- true if releasing jump in the air, enable implementing jumping special abilities manual
    
    --The current transformation mode of jazz
    if(player.tf_mode == nil) then
        player.tf_mode = TF_MODES.BOT_MODE -- The default jazz mode is botmode
    end
    player.transformation_tics = TRANFORMATION_TICS_MAX --Transformation is allowed only when tics <= 0 (so this is its cooldown)
    player.bullet_tics = 0 --Current bullet tic countdown

    player.special_boost_tics = 0 -- Current tic countdown, applying special speed boost while tic counter is above 0

    player.sound_tics_idle = MAX_IDLE_SOUND_TICS --Sound countdowns
    player.sound_tics_mid = MAX_MID_SOUND_TICS --Sound countdowns
    player.sound_tics_high = MAX_HIGH_SOUND_TICS --Sound countdowns
    
    --The angle of car steering
    player.mo.steering = 0 
    player.mo.applied_speed = 0
    player.mo.applied_angle = 0
    player.forward_input = 0
    player.braking_input = 0

    -- True if the jazz was on the ground in the previous frame
    player.mo.prevcarground = true

    --List of bosses found on this level
    player.bosses_detected = {}

    return true
end)


--Switch off the jazz skin
--Returns true if could switch offset jazz, false if player wan't jazz in the first place
rawset(_G, "SwichOff", function(player)
	if(Valid(player.mo)) then
		return false
	end


    if(player.tf_mode == TF_MODES.CAR_MODE) then
        player.mo.state = S_TF_TO_BOT
    end

    --Remove controls
	player.jumpheld = nil
	player.spinheld = nil
    player.forwardheld = nil
	player.prevjumpheld = nil
	player.prevspinheld = nil

	player.mo.prevmomz = nil
	player.jumpspecial = nil

    player.tf_mode = nil
    player.transformation_tics = nil
    
    player.mo.steering = nil

    player.mo.applied_speed = nil
    player.mo.applied_angle = nil
    player.forward_input = nil
    player.braking_input = nil

    player.mo.prevcarground = nil
    player.mo.prevskin = nil

    return true
end)


--Adds time to the special boost ability
local function AddSpecialBoostTics(player, tics)
    if(player.special_boost_tics == nil) then
        player.special_boost_tics = 0
    end
    player.special_boost_tics = clamp($ + tics, 0, SPECIAL_BOOST_TICS_MAX) 
end

--Be able to break obstacles and power up springs by reaching a certain speed limit
local function PowerUp(player, limit)
    --Break
    local speed = GetRealSpeed(player)
    if(speed > limit) then
        player.powers[pw_strong] = $|STR_BUST|STR_SPRING|STR_SPIKE
    else
        player.powers[pw_strong] = $&~STR_BUST&~STR_SPRING&~STR_SPIKE
    end

end

------------------------------------------------------------
------------------ All Sound definitions -------------------
------------------------------------------------------------


------------------------------ Transformation sounds
-- Transformion to car
sfxinfo[sfx_tf_car] = {
	singular = false,
	priority = 64
}

-- Transformion to bot
sfxinfo[sfx_tf_bot] = {
	singular = false,
	priority = 64
}
------------------------------ Car sounds
-- Starting the engine
sfxinfo[sfx_rev] = {
	singular = false,
	priority = 65
}
-- No movement
sfxinfo[sfx_idle] = {
	singular = false,
	priority = 65
}
-- Medium speed
sfxinfo[sfx_mid] = {
	singular = false,
	priority = 65
}

-- High to max speed
sfxinfo[sfx_high] = {
	singular = false,
	priority = 65
}
-- Braking sound
sfxinfo[sfx_brake] = {
    singular = false,
	priority = 64
}
-- Bumping sound
sfxinfo[sfx_bump] = {
    singular = false,
	priority = 64
}
-- Horn sound
sfxinfo[sfx_horn] = {
    singular = false,
	priority = 66
}
------------------------------ Bot sounds
-- Sound of slug bullet firing
sfxinfo[sfx_bullet] = {
	singular = false,
	priority = 64
}
-- Sound of bullet hitting an enemy with a bullet
sfxinfo[sfx_hit] = {
	singular = false,
	priority = 64
}
-- Sound of bullet hitting a wall with a bullet
sfxinfo[sfx_miss] = {
    singular = false,
	priority = 64
}
--Sounds of crescent blast firing 
sfxinfo[sfx_blast] = {
    singular = false,
	priority = 64
}
sfxinfo[sfx_expld] = {
    singular = false,
	priority = 64
}
-- Sound of landing on the ground
sfxinfo[sfx_land] = {
	singular = false,
	priority = 64
}
-- Sound of climbing on the wall
sfxinfo[sfx_climb] = {
	singular = false,
	priority = 64
}
------------------------------ Voicelines sounds
--Starting a boss fight
sfxinfo[sfx_begn01] = {
    singular = false,
	priority = 64
}
--Starting a boss fight
sfxinfo[sfx_begn02] = {
    singular = false,
	priority = 64
}
--Starting a boss fight
sfxinfo[sfx_begn03] = {
    singular = false,
	priority = 64
}
--Starting a boss fight
sfxinfo[sfx_begn04] = {
    singular = false,
	priority = 64
}
--Starting a boss fight
sfxinfo[sfx_begn05] = {
    singular = false,
	priority = 64
}
--Winning a boss fight
sfxinfo[sfx_win01] = {
    singular = false,
	priority = 64
}
--Winning a boss fight
sfxinfo[sfx_win02] = {
    singular = false,
	priority = 64
}
--Winning a boss fight
sfxinfo[sfx_win03] = {
    singular = false,
	priority = 64
}

--Winning a boss fight
sfxinfo[sfx_win04] = {
    singular = false,
	priority = 64
}

--Winning a boss fight
sfxinfo[sfx_win05] = {
    singular = false,
	priority = 64
}



--------------------------------------------------------------------
------------------------ Important Thinkers ------------------------
--------------------------------------------------------------------


--The Base Thinker that plays before others,
--Mostly used to record players input
addHook("PreThinkFrame", function()
	for player in players.iterate() do
        
        --Skin switching behavior
        if(player.mo.prevskin == "jazz" and player.mo.prevskin ~= player.mo.skin) then
            SwichOff(player)
        elseif(player.mo.skin == "jazz" and player.mo.prevskin == nil) then
            SwitchTo(player)
        end
        
        --Prevent any non-jazz players to execute behavior afte rit
        -- continue -- Try find the next jazz player if current is not jazz
        if(player.mo.skin == "jazz") then
        
            -- Prevent drowning in liquid and suffocating in space
            player.powers[pw_underwater] = 0
            player.powers[pw_spacetime] = 0

            -- Controls 
            --Jump button
            if(player.cmd.buttons & BT_JUMP) then
                player.jumpheld = $+1
            else
                player.jumpheld = 0
            end

            --Spin button
            if(player.cmd.buttons & BT_SPIN) then
                player.spinheld = $+1
            else
                player.spinheld = 0
            end

            --Forward button (accounting for joystick too)
            if(player.cmd.forwardmove >= 40) then
                player.forwardheld = $+1
            else
                player.forwardheld = 0
            end

            --Jump special ability (available when letting go of jump button in the air)
            if(not P_IsObjectOnGround(player.mo) and player.jumpheld == 0) then
                player.jumpspecial = 1
            elseif(player.mo.eflags & MFE_JUSTHITFLOOR or player.powers[pw_carry] ~= 0) then
                player.jumpspecial = 0
            end
        end
    end
end)





-- Regular behaviors for both modes (such as Passive Skills)
addHook("PlayerThink", function(player)
    
    if(not Valid(player.mo, "jazz")) then
        return 
    end

    if(player.special_boost_tics > 0) then --Special speed boost is in the affect
        player.special_boost_tics = $-1 --Counting down special speed boost timer
    end

    --Try to 
    PowerUp(player, BREAK_SPEED)
    
    --Add special time if using a spring
    if(player.powers[pw_justsprung] ~= 0) then
        AddSpecialBoostTics(player, SPECIAL_BOOST_SPRING_TICS)
    end
    
    local radius = 1000*FRACUNIT

    --Find boss nearby
    searchBlockmap("objects", function(jazzmo, foundmo)
        if(foundmo.valid and foundmo.flags & MF_BOSS == 0) then
            return -- Stop if boss is not actually a boss
        end
        --Make sure that this boss wasn't detected
        for index, boss in ipairs(jazzmo.player.bosses_detected) do
            if boss == foundmo then
                return true
            end
        end
        --Add the boss to the list of detected bossed on this level
        table.insert(jazzmo.player.bosses_detected, foundmo)
        if(not S_SoundPlaying(jazzmo, sfx_begn01)) then
            S_StartSound(jazzmo, P_RandomRange(sfx_begn01,sfx_begn05)) --Play sound when found 
        end
    end, player.mo,player.mo.x-radius, player.mo.x+radius, player.mo.y-radius, player.mo.y+radius)
    
end)



--The Thinker that plays after other thikers,
--mostly used to clean up, record the previous state, 
--and jump and spin button holding
addHook("PostThinkFrame", function()
	for player in players.iterate() do
		if(Valid(player.mo, "jazz")) then
            

            -- Prevent changing states
            -- if (player.mo.state ~= S_CAR_MODE and player.mo.state ~= S_TF_TO_BOT) then 
            if(player.tf_mode == TF_MODES.CAR_MODE and player.mo.state ~= S_TF_TO_BOT and player.mo.state ~= S_TF_TO_CAR) then
                player.mo.state = S_CAR_MODE
            end
            
            --Keep player in shooting stance when landing 
            if(player.mo.prevstate == S_WEAPON_STANCE and player.mo.eflags & MFE_JUSTHITFLOOR)then
                player.mo.state = S_WEAPON_STANCE
            end

			player.mo.prevskin = player.mo.skin
            player.mo.prevstate = player.mo.state

			player.prevjumpheld = player.jumpheld
			player.prevspinheld = player.spinheld


            --Recording if the car was on the ground in previous tic
            player.mo.prevcarground = P_IsObjectOnGround(player.mo)
		end
	end
end)

addHook("MobjDamage", function(target, inflictor, source, damagetype)

    if(not Valid(source, "jazz")) then
        return --Target wasn't killed by Jazz
    end

    --Jazz killed the boss
    if(target.flags & MF_ENEMY) then
        AddSpecialBoostTics(source.player, SPECIAL_BOOST_DMG_TICS)
    end
end)

addHook("MobjDeath", function(target, inflictor, source, damagetype)

    if(not Valid(source, "jazz")) then --Target wasn't killed by Jazz
        return 
    end

    if(target.flags & MF_BOSS ~= 0) then --Jazz killed the boss
        --Play sound
        if(not S_SoundPlaying(source, sfx_win01)) then
            S_StartSoundAtVolume(source, P_RandomRange(sfx_win01,sfx_win05), VOICE_VOLUME)
        end
    elseif(target.type == MT_RING) then --Jazz collected a ring
        --Adding tics to the boost timer, making sure it doesn't go aboe the limit
        AddSpecialBoostTics(source.player, SPECIAL_BOOST_RING_TICS)
    elseif(target.flags & MF_MONITOR) then --Jazz collected a ring
        --Adding tics to the boost timer, making sure it doesn't go aboe the limit
        AddSpecialBoostTics(source.player, SPECIAL_BOOST_MONITOR_TICS)
    end
end)


-------------------------------------------------------------
------------------------ All Actions ------------------------
-------------------------------------------------------------

local function A_BLAST_EXPLODE(actor, var1, var2)
    --Explosion effect
    actor.destscale = 10*FRACUNIT
    actor.scalespeed = FRACUNIT
    actor.height = actor.destscale
    actor.radius = actor.destscale

    local dust = P_SpawnMobjFromMobj(actor, 0, 0, 0, MT_BDST)
    dust.destscale = actor.destscale * 2
    dust.scalespeed = actor.scalespeed * 2
end

local function A_BULLET_STANCE(actor, var1, var2)
    if(not Valid(actor, "jazz")) then
        return false --Wasn't called for jazz
    end

    -- actor.player.bullet_tics = BULLET_TICS_MAX
end

--Start climbing
local function A_CLIMBING(actor, var1, var2) 
    if(not Valid(actor, "jazz")) then
        return false --Wasn't called for jazz
    end
end

-- Sets up jazz for transformation into the bot
local function A_TF_TO_BOT(actor, var1, var2)
    -- print("to bot!")
    -- if(not Valid(actor, "jazz")) then
    --     return false --Wasn't called for jazz
    -- end
    --Restore mouse input
    input.setMouseGrab(true)

    -- Restore the original height
    actor.player.height = skins[actor.player.skin].height
    actor.player.shieldscale = skins[actor.skin].shieldscale

    -- Change the current transformation state
    actor.player.tf_mode = TF_MODES.BOT_MODE
    actor.player.transformation_tics = TRANFORMATION_TICS_MAX -- resetting transformation tics

    --Breaking floors
    actor.player.powers[pw_strong] = $|STR_FLOOR

    actor.player.charflags = $&~SF_NOSKID

    -- Artificial jump while transforming 
    if((actor.player.jumpheld >= 1 or input.gameControlDown(GC_FIRE) ~= 0) and P_IsObjectOnGround(actor) and actor.eflags & MFE_UNDERWATER == 0) then
        P_SetObjectMomZ(actor, TO_BOT_JUMP, true)
       actor.player.jumpspecial = 1 -- Count as a jump
       --Play jumping animation
       states[actor.state].nextstate = S_PLAY_JUMP
    --Keep walking 
    elseif(actor.player.speed > FRACUNIT) then -- At low speed on the ground
        --Play walking animation
        states[actor.state].nextstate = S_PLAY_WALK
    --Transforming at standstill
    else
        --Play normal animations
        states[actor.state].nextstate = S_PLAY_STND
    end
    
    -- Play sound
    S_StartSound(actor, sfx_tf_bot)
end


-- Sets up jazz for transformation into the car
local function A_TF_TO_CAR(actor, var1, var2)
    if(not Valid(actor, "jazz")) then
        return false --Wasn't called for jazz
    end

    --Ignore mouse input
    input.setMouseGrab(false)

    -- Reduce the height of car to fit into the spingap
    actor.player.height = actor.player.spinheight
    actor.player.shieldscale = CAR_SHIELD_SCALE

    --Stop breaking floors
    actor.player.powers[pw_strong] = $&~STR_FLOOR

    actor.player.charflags = $|SF_NOSKID

    -- Change the current transformation state
    actor.player.tf_mode = TF_MODES.CAR_MODE
    actor.player.transformation_tics = TRANFORMATION_TICS_MAX -- resetting transformation tics

    -- Play sound
    S_StartSound(actor, sfx_tf_car)
end



------------------------------------------------------------
------------------ All State definitions -------------------
------------------------------------------------------------


--State of shooting bullets 
states[S_WEAPON_STANCE] = {
	sprite = SPR_PLAY,
	frame = SPR2_BUSN,
	tics = TICRATE,
    action = A_BULLET_STANCE,
    nextstate = S_PLAY_STND
}
states[S_CLIMBING] = {
    sprite = SPR_PLAY,
	frame = SPR2_RIDE,
	tics = -1,
    action = A_CLIMBING,
    nextstate = S_PLAY_JUMP
}
--State of car
states[S_CAR_MODE] = {
	sprite = SPR_PLAY,
	frame = SPR2_CAR_,
	tics = -1
}

-- Transformation sequence into the car mode
states[S_TF_TO_CAR] = {
	sprite = SPR_PLAY,
	frame = SPR2_TFCA|FF_ANIMATE|A,
	tics = 2*(TICRATE/5),
	action = A_TF_TO_CAR,
    var1 = 2,
    var2 = TICRATE/5,
    nextstate = S_CAR_MODE
}

-- Transformation sequence into the bot mode
states[S_TF_TO_BOT] = {
	sprite = SPR_PLAY,
	frame = SPR2_TFBO|FF_ANIMATE|A,
	tics = states[S_TF_TO_CAR].tics,
	action = A_TF_TO_BOT,
    var1 = states[S_TF_TO_CAR].var1,
    var2 = states[S_TF_TO_CAR].var2,
	nextstate = S_PLAY_STND
}



-- Projectile: Blast that is shot
states[S_BLAST_SHOT] = {
	sprite = SPR_BLST,
    frame = A,
	tics = 10*TICRATE
}

-- Projectile: The explosion of the blast
states[S_BLAST_EXPLODE] = {
	sprite = SPR_BLST,
    frame = A|FF_TRANS50,
    action = A_BLAST_EXPLODE,
	tics = TICRATE/3
}

-- Projectile: The dust explosion of the blast
states[S_BLAST_DUST] = {
	sprite = SPR_BLST,
    frame = A|FF_TRANS50|FF_FULLDARK,
	tics = states[S_BLAST_EXPLODE].tics/2
}


-- Projectile: Bullet that is shot
states[S_BULL_SHOT] = {
	sprite = SPR_BULL,
    frame = A,
	tics = TICRATE
}

-- The regular tire trace 
states[S_TRCK] = {
	sprite = SPR_TRCK,
    frame = A,
	tics = 10*TICRATE,
    nextstate = S_NULL
}

-- The regulat tire trace 
states[S_FIRE] = {
	sprite = SPR_SFLM,
    frame = A,
	tics = 10,
    nextstate = S_NULL
}


------------------------------------------------------------
----------------- All Objects definitions ------------------
------------------------------------------------------------



--Tire Traces
mobjinfo[MT_TRCK] = {
    spawnstate = S_TRCK,
    radius = 13*FRACUNIT,
    height = 7*FRACUNIT,
    flags = mobjinfo[MT_SPINDUST].flags
}

--Tire fire 
mobjinfo[MT_FIRE] = {
    spawnstate = S_FIRE,
    radius = 13*FRACUNIT,
    height = 7*FRACUNIT,
    flags = mobjinfo[MT_TRCK].flags|MF_FIRE
}

--The dust effect after Crescent explosion
mobjinfo[MT_BDST] = {
    spawnstate = S_BLAST_DUST,
    flags = MF_SCENERY|MF_NOGRAVITY
}


--Heavy blast
mobjinfo[MT_BLST] = {
    spawnstate = S_BLAST_SHOT,
    deathstate = S_BLAST_EXPLODE,
    deathsound = sfx_expld,
    speed = 90*FRACUNIT,
    radius = 16*FRACUNIT,
    height = 32*FRACUNIT,
    damage = 2,
    flags = MF_MISSILE|MF2_SUPERFIRE
}

--[[
--Explosion caused by the blast
mobjinfo[MT_EXPL] = {
    spawnstate = S_BLAST_SHOT,
    speed = 70*FRACUNIT,
    radius = 32*FRACUNIT,
    height = 32*FRACUNIT,
    damage = 2,
    flags = MF_MISSILE|MF_GRENADEBOUNCE|MF_NOGRAVITY
}
]]--


--A bullet
mobjinfo[MT_BULL] = {
    spawnstate = S_BULL_SHOT,
    speed = 100*FRACUNIT,
    radius = 10*FRACUNIT,
    height = 10*FRACUNIT,
    damage = 1,
    flags = MF_MISSILE|MF_NOGRAVITY
}